/*
 * Useless Machine - a simple, randomized implementation of useless machine.
 * Copyright (C) 2013 Colin Leroy <colin@colino.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 * 
 */

#include <Servo.h>
#include <avr/power.h>
#include <avr/sleep.h>

int button = 2;
int buttonInterrupt = button -2;

/* definition de nos variables */
int servoRelay = 3;
int openBoxServoPin = 4;
int armServoPin = 5;

#define RELAY_CLOSE HIGH
#define RELAY_OPEN LOW

#define OFF RELAY_OPEN
#define ON RELAY_CLOSE

#define OPEN_BOX 0
#define ARM  1

#define SERVO_UP -1
#define SERVO_DOWN +1

Servo servo[2];
int servoStartAngle[2];
int servoEndAngle[2];
int servoCurAngle[2];
int lastButtonState = -1;

/* récupère l'angle du servo numéro 'numServo' (normalisé) */
int readServo(int numServo) {
  return servo[numServo].read() - 90;
}

/* Envoie un nouvel angle 'angle' (normalisé) au servo 'numServo'
 * en vérifiant que la valeur demandée est dans les clous
 */
boolean setServo(int numServo, int angle) {
  if (angle > servoStartAngle[numServo]) {
    return false;
  }
  if (angle < servoEndAngle[numServo]) {
    return false;
  }
  if (readServo(numServo) == angle) {
    return false;
  }

  int realAngle = angle + 90;
  servo[numServo].write(realAngle);

  return true;
}

/* Deplace le servo 'numServo' dans la direction 'direction' */
boolean moveServo(int numServo, int direction) {
  int curAngle = readServo(numServo);
  int newAngle = curAngle + direction;

  if (newAngle > servoStartAngle[numServo] && direction > 0) {
    newAngle = servoStartAngle[numServo];
  }

  if (newAngle < servoEndAngle[numServo] && direction < 0) {
    newAngle = servoEndAngle[numServo];
  }

  return setServo(numServo, newAngle);
}

/* initialise les variables d'un servo (numéro, pin de controle,
 * angles min et max.
 */
void initServo(int numServo, int controlPin,
               int startAngle, int endAngle) {
  servoStartAngle[numServo] = startAngle;
  servoEndAngle[numServo]   = endAngle;
  servo[numServo].attach(controlPin);
  
  Serial.print("Servo ");
  Serial.print(numServo);
  Serial.print(" initialized from ");
  Serial.print(startAngle);
  Serial.print(" to ");
  Serial.println(endAngle);

  setServo(numServo, startAngle);
}

/* allume ou éteint les servos */
void powerServo(int state) {
  digitalWrite(servoRelay, state);
}

/* Au démarrage */
void setup() {
  Serial.begin(9600);

  /* on initialise le générateur de nombres aléatoires */
  randomSeed(analogRead(0));

  pinMode(13, OUTPUT);
  digitalWrite(13, LOW);

  pinMode(button, INPUT);

  /* on éteint les servos */
  powerServo(OFF);
  pinMode(servoRelay, OUTPUT);
}

void wakeUp() {
  /* do nothing */
}

/* économies d'énergie lorsque l'Arduino est en veille */
void sleepWait() {
  attachInterrupt(buttonInterrupt, wakeUp, LOW);
  set_sleep_mode(SLEEP_MODE_PWR_DOWN);
  sleep_enable();
  power_adc_disable();
  power_spi_disable();
  power_timer0_disable();
  power_timer1_disable();
  power_timer2_disable();
  power_twi_disable();
  sleep_mode();
  
  sleep_disable();
  power_all_enable();
  detachInterrupt(buttonInterrupt);
  delay (100);
}

/* Ouverture de la boite */
void openBox(int noCloseBack) {
  /* définition de la façon d'ouvrir, au hasard
  /* random (a, b) renvoie un nombre entre a et b.
   * Lorsqu'on utilise random (a - b) - c, cela permet
   * d'avoir un nombre aléatoire entre a-c et b-c,
   * a-c étant négatif pour les plus petites valeurs de a.
   * Cela permet de faire un nombre aléatoire entre 0 et b - c, 
   * avec plus de chances de tomber sur 0 (étant donné que tout
   * nombre aléatoire inférieur à c donnera un résultat inférieur
   * à zéro.

  /* combien de pauses
   * random(0, 5) - 2 permet de tirer -2, -1, 0, 1, 2, 3. C'est à dire
   * qu'on a 3 chances sur 6 de ne faire aucune pause. */
  int numPause = random(0, 5) - 2;
  /* quelle vitesse de déplacement */
  int servoSpeed = random(1, 10);
  /* Durée des pauses (en ms) */
  int servoPause = random(1,3) * 100;
  /* Retours en arrière */
  int numBack = (random(0,10) - 6) * 2;
  /* Est-ce qu'on referme ? */
  int closeBack = noCloseBack ? 0 : (random(0, 10) - 8);
  int fast = random(1, 2);
  int i;
  int travelled = 0;
  
  servoSpeed *= fast;

  /* remise à zéro des nombres aléatoires tirés < 0 */
  if (numBack < 0)
    numBack = 0;
 
  if (numPause < 0)
    numPause = 0;
 
  /* on commence par les hésitations ouvert/fermé */
  for (i = 0; i < numBack; i++) {
    setServo(OPEN_BOX, servoStartAngle[OPEN_BOX] - 20);
    delay(60);
    setServo(OPEN_BOX, servoStartAngle[OPEN_BOX]);
    delay(80);
  }

  /* On ouvre la boite */
  while (moveServo(OPEN_BOX, SERVO_UP*servoSpeed)) {
    /* calcul de l'angle déjà effectué */
    travelled += servoSpeed;
    
    /* On fait les pauses prévues à des angles de 20° minimum */
    if (numPause > 0 && travelled > 19) {
      travelled = 0;
      /* durée de la pause */
      delay(servoPause * 10);
      numPause--;
      /* si on a tiré un refermage, on referme puis on recommence
       * en spécifiant qu'on n'a plus le droit de refermer. */
      if (numPause == 0 && closeBack > 0) {
        closeBox(true, true);
        delay(2000);
        openBox(true);
        return;
      }
    } else {
      /* micro pause (pour ralentir) */
      delay(servoPause);
    }
  }
}

/* sortie du bras qui éteint */
void extendArm() {
  /* Définition des variables différentes */
  /* vitesse */
  int servoSpeed = random(1, 10) * 2;
  /* Pause "timidité" */
  int pauseTimidly = random(0, 10) - 5;
  /* Pause "suspense" à la fin du mouvement */
  int pauseSuspense = random(0, 10) - 5;
  /* On insiste sur le bouton ? */
  int hammer = random(0, 10) - 7;
  
  /* Tant que le déplacement fonctionne et que l'interrupteur est allumé... */
  while (moveServo(ARM, SERVO_UP * servoSpeed) && digitalRead(button) == LOW) {
    /* Au début du mouvement, on fait la pause "timidité" */
    if (pauseTimidly > 0 && readServo(ARM) > 0 && readServo(ARM) < 20) {
      delay(500 * random(1, 4));
      pauseTimidly = 0;
    }
    /* à la fin, la pause suspense */
    if (pauseSuspense > 0 && readServo(ARM) < -10) {
      delay(500 * random(1, 4));
      pauseSuspense = 0;
      if (servoSpeed > 2)
        servoSpeed /= 2;
    }
    delay(60);
  }
}

/* on replie le bras */
void foldArm() {
  boolean buttonWasOff = (digitalRead(button) == HIGH);
  /* tant qu'on peut rétracter et que l'interrupteur est off...
   * (ou n'était pas off dès le début... */
  while (moveServo(ARM, SERVO_DOWN) && (digitalRead(button) == HIGH || !buttonWasOff))
    delay(20);
}

/* Fermeture de la boite */
void closeBox(int fast, boolean force) {
  /* Si l'interrupteur est allumé et qu'on n'est pas en mode forcé,
   * on arrete de fermer */
  if (digitalRead(button) == LOW && !force)
    return;

  if (!fast) {
    int servoSpeed = random(1, 20);
    while (moveServo(OPEN_BOX, SERVO_DOWN * servoSpeed) && digitalRead(button) == HIGH)
      delay(20);
  }

  delay(200);
}

/* fonction principale */
void loop() {
  Serial.println("Sleeping");
  delay(100);

  /* on entre en économie d'énergie tant qu'il ne se passe rien */
  sleepWait();
  Serial.println("Woke up");

  /* On initialise les servos */
  initServo(OPEN_BOX, openBoxServoPin, -14, -45);
  initServo(ARM, armServoPin, 48, -30);

  /* On les allume */
  powerServo(ON);

  /* tant que l'interrupteur est allumé... */
  while (digitalRead(button) == LOW) {
    /* on ouvre la boite... */
    openBox(false);
    /* on sort le bras... */
    extendArm();
    /* on range le bras... (cette fonction s'arrete si l'utilisateur
     * rallume l'interrupteur ! le robot ne se laisse pas faire !) */
    foldArm();
    /* on ferme la boite... (cette fonction s'arrete si l'utilisateur
     * rallume l'interrupteur ! le robot ne se laisse pas faire !) */
    closeBox(false, false);
    /* si l'inter a été rallumé, on va réentrer dans cette boucle. */
  }

  /* on ré-éteint les servos et au prochain loop, on se mettra en veille */
  powerServo(OFF);
}

